@@ -3,13 +3,14 @@
  * Copyright (c) 2004-2006 The Trustees of Indiana University and Indiana
  *                         University Research and Technology
  *                         Corporation.  All rights reserved.
- * Copyright (c) 2004-2007 The University of Tennessee and The University
+ * Copyright (c) 2004-2008 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
  * Copyright (c) 2004-2006 High Performance Computing Center Stuttgart,
  *                         University of Stuttgart.  All rights reserved.
  * Copyright (c) 2004-2006 The Regents of the University of California.
  *                         All rights reserved.
+ * Copyright (c) 2007      Cisco Systems, Inc.  All rights reserved.
  * $COPYRIGHT$
  *
  * Additional copyrights may follow
@@ -18,9 +19,13 @@
  */
 
 #include "ompi_config.h"
+
+#include <stdio.h>
+
 #include "ompi/datatype/datatype.h"
 #include "ompi/datatype/datatype_internal.h"
 #include "ompi/datatype/convertor_internal.h"
+#include <stdio.h>
 
 #if OMPI_ENABLE_DEBUG
 #include "opal/mca/base/mca_base_param.h"
@@ -40,7 +45,7 @@ int ompi_ddt_dfd = -1;
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,             \
             0, 0, 0, 0, 0, 0, 0, 0, 0 }
 #define EMPTY_DATA(NAME) NULL, 0, "MPI_" # NAME, {0, 0, NULL}, {0, 0, NULL}, NULL, NULL, ZERO_DDT_ARRAY
-#define BASEOBJ_DATA { OBJ_CLASS(ompi_datatype_t), 1 }
+#define BASEOBJ_DATA OPAL_OBJ_STATIC_INIT(ompi_datatype_t)
 
 /* Using this macro implies that at this point not all informations needed
  * to fill up the datatype are known. We fill them with zeros and then later
@@ -170,6 +175,32 @@ OMPI_DECLSPEC ompi_datatype_t ompi_mpi_2cplex = INIT_BASIC_TYPE( DT_2COMPLEX, 2C
 OMPI_DECLSPEC ompi_datatype_t ompi_mpi_2dblcplex = INIT_BASIC_TYPE( DT_2DOUBLE_COMPLEX, 2DOUBLE_COMPLEX );
 OMPI_DECLSPEC ompi_datatype_t ompi_mpi_unavailable = INIT_UNAVAILABLE_DATA( UNAVAILABLE );
 
+#if OMPI_HAVE_FORTRAN_LOGICAL1
+OMPI_DECLSPEC ompi_datatype_t ompi_mpi_logical1 = INIT_BASIC_FORTRAN_TYPE( DT_LOGIC, LOGICAL1, OMPI_SIZEOF_FORTRAN_LOGICAL1, OMPI_ALIGNMENT_FORTRAN_LOGICAL1, 0 );
+#else
+OMPI_DECLSPEC ompi_datatype_t ompi_mpi_logical1 = INIT_UNAVAILABLE_DATA( LOGICAL1 );
+#endif  /* OMPI_HAVE_FORTRAN_LOGICAL1 */
+#if OMPI_HAVE_FORTRAN_LOGICAL2
+OMPI_DECLSPEC ompi_datatype_t ompi_mpi_logical2 = INIT_BASIC_FORTRAN_TYPE( DT_LOGIC, LOGICAL2, OMPI_SIZEOF_FORTRAN_LOGICAL2, OMPI_ALIGNMENT_FORTRAN_LOGICAL2, 0 );
+#else
+OMPI_DECLSPEC ompi_datatype_t ompi_mpi_logical2 = INIT_UNAVAILABLE_DATA( LOGICAL2 );
+#endif  /* OMPI_HAVE_FORTRAN_LOGICAL2 */
+#if OMPI_HAVE_FORTRAN_LOGICAL4
+OMPI_DECLSPEC ompi_datatype_t ompi_mpi_logical4 = INIT_BASIC_FORTRAN_TYPE( DT_LOGIC, LOGICAL4, OMPI_SIZEOF_FORTRAN_LOGICAL4, OMPI_ALIGNMENT_FORTRAN_LOGICAL4, 0 );
+#else
+OMPI_DECLSPEC ompi_datatype_t ompi_mpi_logical4 = INIT_UNAVAILABLE_DATA( LOGICAL4 );
+#endif  /* OMPI_HAVE_FORTRAN_LOGICAL4 */
+#if OMPI_HAVE_FORTRAN_LOGICAL8
+OMPI_DECLSPEC ompi_datatype_t ompi_mpi_logical8 = INIT_BASIC_FORTRAN_TYPE( DT_LOGIC, LOGICAL8, OMPI_SIZEOF_FORTRAN_LOGICAL8, OMPI_ALIGNMENT_FORTRAN_LOGICAL8, 0 );
+#else
+OMPI_DECLSPEC ompi_datatype_t ompi_mpi_logical8 = INIT_UNAVAILABLE_DATA( LOGICAL8 );
+#endif  /* OMPI_HAVE_FORTRAN_LOGICAL8 */
+
+#if OMPI_HAVE_FORTRAN_REAL2
+OMPI_DECLSPEC ompi_datatype_t ompi_mpi_real2 = INIT_BASIC_FORTRAN_TYPE( DT_FLOAT, REAL2, OMPI_SIZEOF_FORTRAN_REAL2, OMPI_ALIGNMENT_FORTRAN_REAL2, DT_FLAG_DATA_FLOAT );
+#else
+OMPI_DECLSPEC ompi_datatype_t ompi_mpi_real2 = INIT_UNAVAILABLE_DATA( REAL2 );
+#endif
 #if OMPI_HAVE_FORTRAN_REAL4
 OMPI_DECLSPEC ompi_datatype_t ompi_mpi_real4 = INIT_BASIC_FORTRAN_TYPE( DT_FLOAT, REAL4, OMPI_SIZEOF_FORTRAN_REAL4, OMPI_ALIGNMENT_FORTRAN_REAL4, DT_FLAG_DATA_FLOAT );
 #else
@@ -180,7 +211,7 @@ OMPI_DECLSPEC ompi_datatype_t ompi_mpi_real8 = INIT_BASIC_FORTRAN_TYPE( DT_DOUBL
 #else
 OMPI_DECLSPEC ompi_datatype_t ompi_mpi_real8 = INIT_UNAVAILABLE_DATA( REAL8 );
 #endif
-#if OMPI_HAVE_FORTRAN_REAL16
+#if OMPI_HAVE_FORTRAN_REAL16 && OMPI_REAL16_MATCHES_C
 OMPI_DECLSPEC ompi_datatype_t ompi_mpi_real16 = INIT_BASIC_FORTRAN_TYPE( DT_LONG_DOUBLE, REAL16, OMPI_SIZEOF_FORTRAN_REAL16, OMPI_ALIGNMENT_FORTRAN_REAL16, DT_FLAG_DATA_FLOAT );
 #else
 OMPI_DECLSPEC ompi_datatype_t ompi_mpi_real16 = INIT_UNAVAILABLE_DATA( REAL16 );
@@ -222,7 +253,7 @@ OMPI_DECLSPEC ompi_datatype_t ompi_mpi_complex16 = INIT_BASIC_FORTRAN_TYPE( DT_C
 #else
 OMPI_DECLSPEC ompi_datatype_t ompi_mpi_complex16 = INIT_UNAVAILABLE_DATA( COMPLEX16 );
 #endif
-#if OMPI_HAVE_FORTRAN_REAL16 && OMPI_HAVE_FORTRAN_COMPLEX32
+#if OMPI_HAVE_FORTRAN_REAL16 && OMPI_HAVE_FORTRAN_COMPLEX32 && OMPI_REAL16_MATCHES_C
 OMPI_DECLSPEC ompi_datatype_t ompi_mpi_complex32 = INIT_BASIC_FORTRAN_TYPE( DT_COMPLEX_LONG_DOUBLE, COMPLEX32, OMPI_SIZEOF_FORTRAN_COMPLEX32, OMPI_ALIGNMENT_FORTRAN_COMPLEX32, DT_FLAG_DATA_COMPLEX );
 #else
 OMPI_DECLSPEC ompi_datatype_t ompi_mpi_complex32 = INIT_UNAVAILABLE_DATA( COMPLEX32 );
@@ -277,7 +308,7 @@ const ompi_datatype_t* ompi_ddt_basicDatatypes[DT_MAX_PREDEFINED] = {
     &ompi_mpi_unavailable
 };
 
-ompi_pointer_array_t *ompi_datatype_f_to_c_table = NULL;
+opal_pointer_array_t ompi_datatype_f_to_c_table;
 
 size_t ompi_ddt_local_sizes[DT_MAX_PREDEFINED];
 
@@ -424,8 +455,9 @@ int32_t ompi_ddt_init( void )
     }
 
     /* Create the f2c translation table */
-    ompi_datatype_f_to_c_table = OBJ_NEW(ompi_pointer_array_t);
-    if (NULL == ompi_datatype_f_to_c_table) {
+    OBJ_CONSTRUCT(&ompi_datatype_f_to_c_table, opal_pointer_array_t);
+    if( OPAL_SUCCESS != opal_pointer_array_init(&ompi_datatype_f_to_c_table,
+                                                0, OMPI_FORTRAN_HANDLE_MAX, 64)) {
         return OMPI_ERROR;
     }
     /* All temporary datatypes created on the following statement will get registered
@@ -489,8 +521,69 @@ int32_t ompi_ddt_init( void )
     ompi_mpi_cxx_ldblcplex.flags |= DT_FLAG_DATA_CPP | DT_FLAG_DATA_COMPLEX;
 #endif  /* HAVE_LONG_DOUBLE */
 
+    /* Optional Fortran LOGICAL types */
+#if OMPI_HAVE_FORTRAN_LOGICAL1
+#if (OMPI_SIZEOF_FORTRAN_LOGICAL1 == SIZEOF_CHAR)
+    DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_logical1, "MPI_LOGICAL1", &ompi_mpi_char );
+#elif (OMPI_SIZEOF_FORTRAN_LOGICAL1 == SIZEOF_SHORT)
+    DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_logical1, "MPI_LOGICAL1", &ompi_mpi_short );
+#elif (OMPI_SIZEOF_FORTRAN_LOGICAL1 == SIZEOF_INT)
+    DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_logical1, "MPI_LOGICAL1", &ompi_mpi_int );
+#else
+#   warning "No proper C type found for LOGICAL1"
+    DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_logical1, "MPI_LOGICAL1", &ompi_mpi_unavailable );
+#endif
+    ompi_mpi_logical1.flags |= DT_FLAG_DATA_FORTRAN;
+#endif /* OMPI_HAVE_FORTRAN_LOGICAL1 */
+
+#if OMPI_HAVE_FORTRAN_LOGICAL2
+#if (OMPI_SIZEOF_FORTRAN_LOGICAL2 == SIZEOF_SHORT)
+    DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_logical2, "MPI_LOGICAL2", &ompi_mpi_short );
+#elif (OMPI_SIZEOF_FORTRAN_LOGICAL2 == SIZEOF_INT)
+    DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_logical2, "MPI_LOGICAL2", &ompi_mpi_int );
+#else
+#   warning "No proper C type found for LOGICAL2"
+    DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_logical2, "MPI_LOGICAL2", &ompi_mpi_unavailable );
+#endif
+    ompi_mpi_logical2.flags |= DT_FLAG_DATA_FORTRAN;
+#endif /* OMPI_HAVE_FORTRAN_LOGICAL2 */
+
+#if OMPI_HAVE_FORTRAN_LOGICAL4
+#if (OMPI_SIZEOF_FORTRAN_LOGICAL4 == SIZEOF_INT)
+    DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_logical4, "MPI_LOGICAL4", &ompi_mpi_int );
+#elif (OMPI_SIZEOF_FORTRAN_LOGICAL4 == SIZEOF_LONG)
+    DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_logical4, "MPI_LOGICAL4", &ompi_mpi_long );
+#else
+#   warning "No proper C type found for LOGICAL4"
+    DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_logical4, "MPI_LOGICAL4", &ompi_mpi_unavailable );
+#endif
+    ompi_mpi_logical4.flags |= DT_FLAG_DATA_FORTRAN;
+#endif /* OMPI_HAVE_FORTRAN_LOGICAL4 */
+
+#if OMPI_HAVE_FORTRAN_LOGICAL8
+#if (OMPI_SIZEOF_FORTRAN_LOGICAL8 == SIZEOF_LONG)
+    DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_logical8, "MPI_LOGICAL8", &ompi_mpi_long );
+#elif (OMPI_SIZEOF_FORTRAN_LOGICAL8 == SIZEOF_LONG_LONG)
+    DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_logical8, "MPI_LOGICAL8", &ompi_mpi_long_long_int );
+#else
+#   warning "No proper C type found for LOGICAL8"
+    DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_logical8, "MPI_LOGICAL8", &ompi_mpi_unavailable );
+#endif
+    ompi_mpi_logical8.flags |= DT_FLAG_DATA_FORTRAN | DT_FLAG_DATA_INT;
+#endif /* OMPI_HAVE_FORTRAN_INTEGER8 */
+
     /* Optional Fortran REAL types */
 
+#if OMPI_HAVE_FORTRAN_REAL2
+#if (OMPI_SIZEOF_FORTRAN_REAL2 == SIZEOF_FLOAT)
+    DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_real2, "MPI_REAL2", &ompi_mpi_float );
+#else
+#   warning "No proper C type found for REAL2"
+    DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_real2, "MPI_REAL2", &ompi_mpi_unavailable );
+#endif
+    ompi_mpi_real2.flags |= DT_FLAG_DATA_FORTRAN | DT_FLAG_DATA_FLOAT;
+#endif /* OMPI_HAVE_FORTRAN_REAL2 */
+
 #if OMPI_HAVE_FORTRAN_REAL4
 #if (OMPI_SIZEOF_FORTRAN_REAL4 == SIZEOF_FLOAT)
     DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_real4, "MPI_REAL4", &ompi_mpi_float );
@@ -612,7 +705,7 @@ int32_t ompi_ddt_init( void )
 #endif /* OMPI_HAVE_FORTRAN_COMPLEX16 */
 
 #if OMPI_HAVE_FORTRAN_COMPLEX32
-#if (OMPI_SIZEOF_FORTRAN_COMPLEX32 == 2*SIZEOF_LONG_DOUBLE)
+#if OMPI_REAL16_MATCHES_C && (OMPI_SIZEOF_FORTRAN_COMPLEX32 == 2*SIZEOF_LONG_DOUBLE)
     DECLARE_MPI_SYNONYM_DDT( &ompi_mpi_complex32, "MPI_COMPLEX32", &ompi_mpi_ldblcplex );
 #else
 #   warning "No proper C type found for COMPLEX32"
@@ -634,7 +727,7 @@ int32_t ompi_ddt_init( void )
 #define MOOG(name)                                                      \
     {                                                                   \
         ompi_mpi_##name.d_f_to_c_index =                                \
-            ompi_pointer_array_add(ompi_datatype_f_to_c_table, &ompi_mpi_##name); \
+            opal_pointer_array_add(&ompi_datatype_f_to_c_table, &ompi_mpi_##name); \
         if( ompi_ddt_number_of_predefined_data < (ompi_mpi_##name).d_f_to_c_index ) \
             ompi_ddt_number_of_predefined_data = (ompi_mpi_##name).d_f_to_c_index; \
     }
@@ -671,6 +764,11 @@ int32_t ompi_ddt_init( void )
     MOOG(2integer);
     MOOG(2cplex);
     MOOG(2dblcplex);
+    MOOG(real2);
+    MOOG(logical1);
+    MOOG(logical2);
+    MOOG(logical4);
+    MOOG(logical8);
 
     /* Now the C types */
 
@@ -706,7 +804,7 @@ int32_t ompi_ddt_init( void )
     MOOG(cxx_ldblcplex);
 
     for( i = 0; i < ompi_mpi_cxx_ldblcplex.d_f_to_c_index; i++ ) {
-        ompi_datatype_t* datatype = (ompi_datatype_t*)ompi_pointer_array_get_item( ompi_datatype_f_to_c_table, i );
+        ompi_datatype_t* datatype = (ompi_datatype_t*)opal_pointer_array_get_item(&ompi_datatype_f_to_c_table, i );
 
         if( (datatype->ub - datatype->lb) == (ptrdiff_t)datatype->size ) {
             datatype->flags |= DT_FLAG_NO_GAPS;
@@ -735,7 +833,7 @@ int32_t ompi_ddt_finalize( void )
     }
 
     /* Get rid of the Fortran2C translation table */
-    OBJ_RELEASE(ompi_datatype_f_to_c_table);
+    OBJ_DESTRUCT(&ompi_datatype_f_to_c_table);
 
 #if defined(VERBOSE)
     if( ompi_ddt_dfd != -1 )
@@ -755,7 +853,7 @@ int32_t ompi_ddt_finalize( void )
 #if OMPI_ENABLE_DEBUG
 /*
  * Set a breakpoint to this function in your favorite debugger
- * to make it stopping on all pack and unpack errors.
+ * to make it stop on all pack and unpack errors.
  */
 int ompi_ddt_safeguard_pointer_debug_breakpoint( const void* actual_ptr, int length,
                                                  const void* initial_ptr,
@@ -772,8 +870,10 @@ int ompi_ddt_safeguard_pointer_debug_breakpoint( const void* actual_ptr, int len
 
 static int _dump_data_flags( unsigned short usflags, char* ptr, size_t length )
 {
-    if( length < 21 ) return 0;
-    snprintf( ptr, 21, "-----------[---][---]" );  /* set everything to - */
+    int index = 0;
+
+    if( length < 22 ) return 0;
+    index = snprintf( ptr, 22, "-----------[---][---]" );  /* set everything to - */
     if( usflags & DT_FLAG_DESTROYED )                ptr[0]  = 'd';
     if( usflags & DT_FLAG_COMMITED )                 ptr[1]  = 'c';
     if( usflags & DT_FLAG_CONTIGUOUS )               ptr[2]  = 'C';
@@ -809,7 +909,7 @@ static int _dump_data_flags( unsigned short usflags, char* ptr, size_t length )
             ptr[17] = 'E'; ptr[18] = 'R'; ptr[19] = 'R'; break;
         }
     }
-    return 21;
+    return index;
 }
 
 static int __dump_data_desc( dt_elem_desc_t* pDesc, int nbElems, char* ptr, size_t length )
