@@ -1,14 +1,15 @@
 /*
- * Copyright (c) 2004-2005 The Trustees of Indiana University and Indiana
+ * Copyright (c) 2004-2007 The Trustees of Indiana University and Indiana
  *                         University Research and Technology
  *                         Corporation.  All rights reserved.
- * Copyright (c) 2004-2005 The University of Tennessee and The University
+ * Copyright (c) 2004-2008 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
  * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
  *                         University of Stuttgart.  All rights reserved.
  * Copyright (c) 2004-2005 The Regents of the University of California.
  *                         All rights reserved.
+ * Copyright (c) 2006      Myricom, Inc.  All rights reserved.
  * $COPYRIGHT$
  * 
  * Additional copyrights may follow
@@ -18,7 +19,7 @@
 
 #include "ompi_config.h"
 #include <string.h>
-#include "opal/util/output.h"
+#include "orte/util/show_help.h"
 #include "opal/util/if.h"
 #include "ompi/mca/pml/pml.h"
 #include "ompi/mca/btl/btl.h"
@@ -61,15 +62,16 @@ mca_btl_gm_module_t mca_btl_gm_module = {
         0, /* max size of first fragment */
         0, /* min send fragment size */
         0, /* max send fragment size */
-        0, /* min rdma fragment size */
-        0, /* max rdma fragment size */
+        0, /* btl_rdma_pipeline_send_length */
+        0, /* btl_rdma_pipeline_frag_size */
+        0, /* btl_min_rdma_pipeline_size */
         0, /* exclusivity */
         0, /* latency */
         0, /* bandwidth */
         0, /* flags */
         mca_btl_gm_add_procs,
         mca_btl_gm_del_procs,
-        mca_btl_gm_register, 
+        NULL, 
         mca_btl_gm_finalize,
         mca_btl_gm_alloc, 
         mca_btl_gm_free, 
@@ -77,6 +79,7 @@ mca_btl_gm_module_t mca_btl_gm_module = {
         mca_btl_gm_prepare_dst,
 #if OMPI_ENABLE_MPI_THREADS || OMPI_ENABLE_PROGRESS_THREADS
         mca_btl_gm_send,
+        NULL, /* send immediate */
         mca_btl_gm_put,
         mca_btl_gm_get,
 #else
@@ -84,7 +87,10 @@ mca_btl_gm_module_t mca_btl_gm_module = {
         mca_btl_gm_put_nl,
         mca_btl_gm_get_nl,
 #endif
-        mca_btl_base_dump
+        mca_btl_base_dump,
+        NULL, /* mpool */
+        mca_btl_gm_register_error_cb,
+        mca_btl_gm_ft_event
     }
 };
 
@@ -157,24 +163,19 @@ int mca_btl_gm_del_procs(struct mca_btl_base_module_t* btl,
     return OMPI_SUCCESS;
 }
 
-
-/**
- * Register callback function to support send/recv semantics
- */
-
-int mca_btl_gm_register(
+/* 
+ *Register callback function for error handling..
+ */ 
+int mca_btl_gm_register_error_cb(
                         struct mca_btl_base_module_t* btl, 
-                        mca_btl_base_tag_t tag, 
-                        mca_btl_base_module_recv_cb_fn_t cbfunc, 
-                        void* cbdata)
+                        mca_btl_base_module_error_cb_fn_t cbfunc)
 {
+    
     mca_btl_gm_module_t* gm_btl = (mca_btl_gm_module_t*) btl; 
-    gm_btl->gm_reg[tag].cbfunc = cbfunc; 
-    gm_btl->gm_reg[tag].cbdata = cbdata; 
+    gm_btl->error_cb = cbfunc; /* stash for later */
     return OMPI_SUCCESS;
 }
 
-
 /**
  * Allocate a segment.
  *
@@ -184,7 +185,10 @@ int mca_btl_gm_register(
 
 mca_btl_base_descriptor_t* mca_btl_gm_alloc(
     struct mca_btl_base_module_t* btl,
-    size_t size)
+    struct mca_btl_base_endpoint_t* endpoint,
+    uint8_t order,
+    size_t size,
+    uint32_t flags)
 {
     mca_btl_gm_module_t* gm_btl = (mca_btl_gm_module_t*) btl; 
     mca_btl_gm_frag_t* frag;
@@ -192,9 +196,18 @@ mca_btl_base_descriptor_t* mca_btl_gm_alloc(
     
     if(size <= btl->btl_eager_limit) { 
         MCA_BTL_GM_FRAG_ALLOC_EAGER(gm_btl, frag, rc); 
+        if(NULL == frag) { 
+            return NULL;
+        }
+        frag->type=MCA_BTL_GM_EAGER;
         frag->segment.seg_len = size;
+        
     } else if(size <=  btl->btl_max_send_size) { 
         MCA_BTL_GM_FRAG_ALLOC_MAX(gm_btl, frag, rc); 
+        if(NULL == frag) { 
+            return NULL;
+        }
+        frag->type=MCA_BTL_GM_SEND;
         frag->segment.seg_len = size; 
     } else { 
         return NULL;
@@ -204,7 +217,8 @@ mca_btl_base_descriptor_t* mca_btl_gm_alloc(
     frag->base.des_src_cnt = 1;
     frag->base.des_dst = NULL;
     frag->base.des_dst_cnt = 0;
-    frag->base.des_flags = 0; 
+    frag->base.des_flags = flags; 
+    frag->base.order = MCA_BTL_NO_ORDER;
     return &frag->base;
 }
 
@@ -213,14 +227,14 @@ mca_btl_base_descriptor_t* mca_btl_gm_alloc(
  * Return a segment
  */
 
-int mca_btl_gm_free(
-    struct mca_btl_base_module_t* btl, 
-    mca_btl_base_descriptor_t* des) 
+int mca_btl_gm_free( struct mca_btl_base_module_t* btl, 
+		     mca_btl_base_descriptor_t* des )
 {
     mca_btl_gm_frag_t* frag = (mca_btl_gm_frag_t*)des;
-    if( NULL != frag->registration ){
-        btl->btl_mpool->mpool_release(btl->btl_mpool, 
-                                      (mca_mpool_base_registration_t*) frag->registration);
+
+    if( NULL != frag->registration ) {
+        btl->btl_mpool->mpool_deregister(btl->btl_mpool, (mca_mpool_base_registration_t*) frag->registration);
+        frag->registration = NULL;
     }
 
     MCA_BTL_GM_FRAG_RETURN(btl, frag); 
@@ -239,140 +253,103 @@ mca_btl_base_descriptor_t* mca_btl_gm_prepare_src(
     struct mca_btl_base_endpoint_t* endpoint,
     struct mca_mpool_base_registration_t* registration,
     struct ompi_convertor_t* convertor,
+    uint8_t order,
     size_t reserve,
-    size_t* size
+    size_t* size,
+    uint32_t flags
 )
 {
-    mca_btl_gm_frag_t* frag;
+    mca_btl_gm_frag_t *frag = NULL;
     struct iovec iov;
     uint32_t iov_count = 1;
     size_t max_data = *size;
-    int32_t free_after;
     int rc;
 
 #if (OMPI_MCA_BTL_GM_HAVE_RDMA_GET || OMPI_MCA_BTL_GM_HAVE_RDMA_PUT)
-    /*
-     * If the data has already been pinned and is contigous than we can
-     * use it in place.
-    */
-    if (NULL != registration && 0 == ompi_convertor_need_buffers(convertor)) {
-        size_t reg_len;
-        MCA_BTL_GM_FRAG_ALLOC_USER(btl, frag, rc);
-        if(NULL == frag){
-            return NULL;
+    if(ompi_convertor_need_buffers(convertor) == false && 0 == reserve) {
+        if(registration != NULL || max_data > btl->btl_max_send_size) {
+            MCA_BTL_GM_FRAG_ALLOC_USER(btl, frag, rc);
+            if(NULL == frag) {
+                return NULL;
+            }
+
+            /*
+             * just assign it something..
+             * we will assign the real value in put/get
+             */
+            frag->type = MCA_BTL_GM_PUT;
+            iov.iov_len = max_data;
+            iov.iov_base = NULL;
+
+
+            ompi_convertor_pack(convertor, &iov, &iov_count, &max_data);
+
+            *size = max_data;
+
+            if(NULL == registration) {
+                rc = btl->btl_mpool->mpool_register(btl->btl_mpool,
+                        iov.iov_base, max_data, 0, &registration);
+                if(OMPI_SUCCESS != rc || NULL == registration) {
+                    MCA_BTL_GM_FRAG_RETURN(btl, frag);
+                    return NULL;
+                }
+                /* keep track of the registration we did */
+                frag->registration = registration;
+            }
+
+            frag->segment.seg_len = max_data;
+            frag->segment.seg_addr.pval = iov.iov_base;
+
+            frag->base.des_src     = &frag->segment;
+            frag->base.des_src_cnt = 1;
+            frag->base.des_dst     = NULL;
+            frag->base.des_dst_cnt = 0;
+            frag->base.des_flags   = flags;
+            frag->base.order       = MCA_BTL_NO_ORDER;
+
+            return &frag->base;
         }
-        iov.iov_len = max_data;
-        iov.iov_base = NULL;
-
-        ompi_convertor_pack(convertor, &iov, &iov_count, &max_data, &free_after);
-                                                                                                    
-        frag->segment.seg_len = max_data;
-        frag->segment.seg_addr.pval = iov.iov_base;
-
-        reg_len = (unsigned char*)registration->bound - (unsigned char*)iov.iov_base + 1;
-        
-        /* bump reference count as so that the registration
-         * doesn't go away when the operation completes
-         */
-        btl->btl_mpool->mpool_retain(btl->btl_mpool, 
-                                     (mca_mpool_base_registration_t*) registration);
-        
-        frag->registration = registration;
-
-    /*
-     * if the data is not already pinned - but the leave pinned option is set,
-     * then go ahead and pin contigous data. however, if a reserve is required 
-     * then we must allocated a fragment w/ buffer space
-    */
-    } else if (max_data > btl->btl_max_send_size && 
-               ompi_convertor_need_buffers(convertor) == 0 &&
-               reserve == 0) {
-
-        mca_mpool_base_module_t* mpool = btl->btl_mpool;
-        MCA_BTL_GM_FRAG_ALLOC_USER(btl, frag, rc);
-        if(NULL == frag){
-            return NULL;
-        }
-        iov.iov_len = max_data;
-        iov.iov_base = NULL;
-
-        ompi_convertor_pack(convertor, &iov, &iov_count, &max_data, &free_after);
-                                                                                                
-        frag->segment.seg_len = max_data;
-        frag->segment.seg_addr.pval = iov.iov_base;
-
-        rc = mpool->mpool_register(
-                                   mpool,
-                                   iov.iov_base,
-                                   max_data,
-                                   0,
-                                   &registration);
-
-        if(rc != OMPI_SUCCESS) {
-            MCA_BTL_GM_FRAG_RETURN(btl,frag);
-            return NULL;
-        }
-
-        frag->registration = registration;
-    } 
-
-    /*
-     * if we aren't pinning the data and the requested size is less
-     * than the eager limit pack into a fragment from the eager pool
-    */
-    else 
+    }
 #endif
 
-    if (max_data+reserve <= btl->btl_eager_limit) {
-                                                                                                    
+    if (max_data + reserve <= btl->btl_eager_limit) {
+        /* the data is small enough to fit in the eager frag and
+         * memory is not prepinned */
         MCA_BTL_GM_FRAG_ALLOC_EAGER(btl, frag, rc);
-        if(NULL == frag) {
-            return NULL;
+        if(frag != NULL) {
+            frag->type = MCA_BTL_GM_EAGER;
         }
-                                                                                                    
-        iov.iov_len = max_data;
-        iov.iov_base = (unsigned char*) frag->segment.seg_addr.pval + reserve;
-                                                                                                    
-        rc = ompi_convertor_pack(convertor, &iov, &iov_count, &max_data, &free_after);
-        *size  = max_data;
-        if( rc < 0 ) {
-            MCA_BTL_GM_FRAG_RETURN(btl, frag);
-            return NULL;
-        }
-        frag->segment.seg_len = max_data + reserve;
     }
 
-    /* 
-     * otherwise pack as much data as we can into a fragment
-     * that is the max send size.
-     */
-    else {
-                                                                                                    
+    if(NULL == frag) {
+        /* the data doesn't fit into eager frag or eger frag is
+         * not available */
         MCA_BTL_GM_FRAG_ALLOC_MAX(btl, frag, rc);
         if(NULL == frag) {
             return NULL;
         }
-        if(max_data + reserve > btl->btl_max_send_size){
+        frag->type = MCA_BTL_GM_SEND;
+        if(max_data + reserve > btl->btl_max_send_size) {
             max_data = btl->btl_max_send_size - reserve;
         }
-        iov.iov_len = max_data;
-        iov.iov_base = (unsigned char*) frag->segment.seg_addr.pval + reserve;
-                                                                                                    
-        rc = ompi_convertor_pack(convertor, &iov, &iov_count, &max_data, &free_after);
-        *size  = max_data;
-                                                                                                    
-        if( rc < 0 ) {
-            MCA_BTL_GM_FRAG_RETURN(btl, frag);
-            return NULL;
-        }
-        frag->segment.seg_len = max_data + reserve;
     }
 
-    frag->base.des_src = &frag->segment;
+    iov.iov_len = max_data;
+    iov.iov_base = (unsigned char*) frag->segment.seg_addr.pval + reserve;
+    rc = ompi_convertor_pack(convertor, &iov, &iov_count, &max_data);
+    if(rc < 0) {
+        MCA_BTL_GM_FRAG_RETURN(btl, frag);
+        return NULL;
+    }
+    *size  = max_data;
+    frag->segment.seg_len  = max_data + reserve;
+    frag->base.des_src     = &frag->segment;
     frag->base.des_src_cnt = 1;
-    frag->base.des_dst = NULL;
+    frag->base.des_dst     = NULL;
     frag->base.des_dst_cnt = 0;
-    frag->base.des_flags = 0;
+    frag->base.des_flags   = flags;
+    frag->base.order       = MCA_BTL_NO_ORDER;
+
     return &frag->base;
 }
 
@@ -396,52 +373,48 @@ mca_btl_base_descriptor_t* mca_btl_gm_prepare_dst(
     struct mca_btl_base_endpoint_t* endpoint,
     struct mca_mpool_base_registration_t* registration,
     struct ompi_convertor_t* convertor,
+    uint8_t order,
     size_t reserve,
-    size_t* size)
+    size_t* size,
+    uint32_t flags)
 {
 #if (OMPI_MCA_BTL_GM_HAVE_RDMA_GET || OMPI_MCA_BTL_GM_HAVE_RDMA_PUT)
     mca_btl_gm_frag_t* frag;
     mca_mpool_base_module_t* mpool = btl->btl_mpool;
-    long lb;
     int rc;
 
     MCA_BTL_GM_FRAG_ALLOC_USER(btl, frag, rc);
     if(NULL == frag) {
         return NULL;
     }
-
-    ompi_ddt_type_lb(convertor->pDesc, &lb);
+    
+    /* 
+     *  we don't know that this is for a PUT,
+     *  but it doesn't matter.. they belong 
+     *  on the same list eventually anyway
+     */      
+    frag->type = MCA_BTL_GM_PUT; 
+    
     frag->segment.seg_len = *size;
-    frag->segment.seg_addr.pval = convertor->pBaseBuf + lb + convertor->bConverted;
+    ompi_convertor_get_current_pointer( convertor, (void**)&(frag->segment.seg_addr.pval) );
 
     frag->base.des_src = NULL;
     frag->base.des_src_cnt = 0;
     frag->base.des_dst = &frag->segment;
     frag->base.des_dst_cnt = 1;
-    frag->base.des_flags = 0;
-    if(NULL != registration) {
-        /* bump reference count as so that the registration
-         * doesn't go away when the operation completes
-         */
-        
-        mpool->mpool_retain(mpool, 
-                           (mca_mpool_base_registration_t*) registration); 
-                
-        frag->registration = registration;
+    frag->base.des_flags = flags;
+    frag->base.order = MCA_BTL_NO_ORDER;
 
-    }  else {
-
-        rc = mpool->mpool_register(
-                                   mpool,
-                                   frag->segment.seg_addr.pval,
-                                   frag->segment.seg_len,
-                                   0,
-                                   &registration);
+    if(NULL == registration) {
+        rc = mpool->mpool_register( mpool,
+				    frag->segment.seg_addr.pval,
+				    frag->segment.seg_len,
+				    0,
+				    &registration );
         if(rc != OMPI_SUCCESS) {
             MCA_BTL_GM_FRAG_RETURN(btl,frag);
             return NULL;
         }
-        
         frag->registration = registration;
     }
     return &frag->base;
@@ -470,11 +443,14 @@ static void mca_btl_gm_send_callback( struct gm_port* port, void* context, gm_st
 {
     mca_btl_gm_frag_t* frag = (mca_btl_gm_frag_t*)context;
     mca_btl_gm_module_t* btl = frag->btl;
+    int btl_ownership = frag->base.des_flags & MCA_BTL_DES_FLAGS_BTL_OWNERSHIP;
 
     switch(status) {
         case GM_TRY_AGAIN:
         case GM_SEND_TIMED_OUT:
+#if GM_API_VERSION > 0x200
         case GM_TIMED_OUT:
+#endif
             /* drop all sends to this destination port */
             gm_drop_sends(
                 btl->port,
@@ -500,7 +476,9 @@ static void mca_btl_gm_send_callback( struct gm_port* port, void* context, gm_st
             OPAL_THREAD_UNLOCK(&mca_btl_gm_component.gm_lock);
             frag->base.des_cbfunc(&btl->super, frag->endpoint, &frag->base, OMPI_SUCCESS);
             OPAL_THREAD_LOCK(&mca_btl_gm_component.gm_lock);
-
+            if( btl_ownership ) {
+                mca_btl_gm_free(&btl->super, &frag->base);
+            }
             /* return the send token and deque pending fragments */
             MCA_BTL_GM_RETURN_TOKEN(btl);
             break;
@@ -516,6 +494,9 @@ static void mca_btl_gm_send_callback( struct gm_port* port, void* context, gm_st
             OPAL_THREAD_UNLOCK(&mca_btl_gm_component.gm_lock);
             frag->base.des_cbfunc(&btl->super, frag->endpoint, &frag->base, OMPI_ERROR);
             OPAL_THREAD_LOCK(&mca_btl_gm_component.gm_lock);
+            if( btl_ownership ) {
+                mca_btl_gm_free(&btl->super, &frag->base);
+            }
             break;
     }
 }
@@ -544,7 +525,6 @@ static int mca_btl_gm_send_nl(
     frag->btl = gm_btl;
     frag->endpoint = endpoint; 
     frag->hdr->tag = tag;
-    frag->type = MCA_BTL_GM_SEND;
 
     /* queue the descriptor if there are no send tokens */
     MCA_BTL_GM_ACQUIRE_TOKEN_NL(gm_btl, frag);
@@ -552,35 +532,33 @@ static int mca_btl_gm_send_nl(
     /* post the send descriptor */
     if(frag->base.des_flags & MCA_BTL_DES_FLAGS_PRIORITY &&  
        frag->size == mca_btl_gm_component.gm_eager_frag_size) {
-        gm_send_with_callback(
-            gm_btl->port,
-            frag->hdr,
-            mca_btl_gm_component.gm_eager_frag_size,
-            frag->segment.seg_len + sizeof(mca_btl_base_header_t),
-            GM_HIGH_PRIORITY,
-            endpoint->endpoint_addr.node_id,
-            endpoint->endpoint_addr.port_id,
-            mca_btl_gm_send_callback,
-            frag);
+        gm_send_with_callback( gm_btl->port,
+			       frag->hdr,
+			       mca_btl_gm_component.gm_eager_frag_size,
+			       frag->segment.seg_len + sizeof(mca_btl_base_header_t),
+			       GM_HIGH_PRIORITY,
+			       endpoint->endpoint_addr.node_id,
+			       endpoint->endpoint_addr.port_id,
+			       mca_btl_gm_send_callback,
+			       frag );
     } else {
-        gm_send_with_callback(
-            gm_btl->port,
-            frag->hdr,
-            mca_btl_gm_component.gm_max_frag_size,
-            frag->segment.seg_len + sizeof(mca_btl_base_header_t),
-            GM_LOW_PRIORITY,
-            endpoint->endpoint_addr.node_id,
-            endpoint->endpoint_addr.port_id,
-            mca_btl_gm_send_callback,
-            frag);
+        gm_send_with_callback( gm_btl->port,
+			       frag->hdr,
+			       mca_btl_gm_component.gm_max_frag_size,
+			       frag->segment.seg_len + sizeof(mca_btl_base_header_t),
+			       GM_LOW_PRIORITY,
+			       endpoint->endpoint_addr.node_id,
+			       endpoint->endpoint_addr.port_id,
+			       mca_btl_gm_send_callback,
+			       frag );
     }
-
     if(opal_list_get_size(&gm_btl->gm_repost)) {
         mca_btl_gm_frag_t* frag;
         while(NULL != (frag = (mca_btl_gm_frag_t*)opal_list_remove_first(&gm_btl->gm_repost))) {
             gm_provide_receive_buffer(gm_btl->port, frag->hdr, frag->size, frag->priority);
         }
     }
+
     return OMPI_SUCCESS;
 }
 
@@ -607,8 +585,7 @@ int mca_btl_gm_send(
     frag->btl = gm_btl;
     frag->endpoint = endpoint; 
     frag->hdr->tag = tag;
-    frag->type = MCA_BTL_GM_SEND;
-
+    
     /* queue the descriptor if there are no send tokens */
     OPAL_THREAD_LOCK(&mca_btl_gm_component.gm_lock);
     MCA_BTL_GM_ACQUIRE_TOKEN(gm_btl, frag);
@@ -616,27 +593,25 @@ int mca_btl_gm_send(
     /* post the send descriptor */
     if(frag->base.des_flags & MCA_BTL_DES_FLAGS_PRIORITY &&  
        frag->size == mca_btl_gm_component.gm_eager_frag_size) {
-        gm_send_with_callback(
-            gm_btl->port,
-            frag->hdr,
-            mca_btl_gm_component.gm_eager_frag_size,
-            frag->segment.seg_len + sizeof(mca_btl_base_header_t),
-            GM_HIGH_PRIORITY,
-            endpoint->endpoint_addr.node_id,
-            endpoint->endpoint_addr.port_id,
-            mca_btl_gm_send_callback,
-            frag);
+        gm_send_with_callback( gm_btl->port,
+			       frag->hdr,
+			       mca_btl_gm_component.gm_eager_frag_size,
+			       frag->segment.seg_len + sizeof(mca_btl_base_header_t),
+			       GM_HIGH_PRIORITY,
+			       endpoint->endpoint_addr.node_id,
+			       endpoint->endpoint_addr.port_id,
+			       mca_btl_gm_send_callback,
+			       frag );
     } else {
-        gm_send_with_callback(
-            gm_btl->port,
-            frag->hdr,
-            mca_btl_gm_component.gm_max_frag_size,
-            frag->segment.seg_len + sizeof(mca_btl_base_header_t),
-            GM_LOW_PRIORITY,
-            endpoint->endpoint_addr.node_id,
-            endpoint->endpoint_addr.port_id,
-            mca_btl_gm_send_callback,
-            frag);
+        gm_send_with_callback( gm_btl->port,
+			       frag->hdr,
+			       mca_btl_gm_component.gm_max_frag_size,
+			       frag->segment.seg_len + sizeof(mca_btl_base_header_t),
+			       GM_LOW_PRIORITY,
+			       endpoint->endpoint_addr.node_id,
+			       endpoint->endpoint_addr.port_id,
+			       mca_btl_gm_send_callback,
+			       frag );
     }
 
     if(opal_list_get_size(&gm_btl->gm_repost)) {
@@ -660,12 +635,15 @@ static void mca_btl_gm_put_callback( struct gm_port* port, void* context, gm_sta
 {
     mca_btl_gm_frag_t* frag = (mca_btl_gm_frag_t*)context;
     mca_btl_gm_module_t* btl = frag->btl;
+    int btl_ownership = frag->base.des_flags & MCA_BTL_DES_FLAGS_BTL_OWNERSHIP;
 
     /* call the completion callback */
     switch(status) {
         case GM_TRY_AGAIN:
         case GM_SEND_TIMED_OUT:
+#if GM_API_VERSION > 0x200
         case GM_TIMED_OUT:
+#endif
             /* drop all sends to this destination port */
 
             gm_drop_sends(
@@ -692,6 +670,9 @@ static void mca_btl_gm_put_callback( struct gm_port* port, void* context, gm_sta
             OPAL_THREAD_UNLOCK(&mca_btl_gm_component.gm_lock);
             frag->base.des_cbfunc(&btl->super, frag->endpoint, &frag->base, OMPI_SUCCESS);
             OPAL_THREAD_LOCK(&mca_btl_gm_component.gm_lock);
+            if( btl_ownership ) {
+                mca_btl_gm_free(&btl->super, &frag->base);
+            }
 
             /* return the send token and deque pending fragments */
             MCA_BTL_GM_RETURN_TOKEN(btl);
@@ -707,6 +688,9 @@ static void mca_btl_gm_put_callback( struct gm_port* port, void* context, gm_sta
             OPAL_THREAD_UNLOCK(&mca_btl_gm_component.gm_lock);
             frag->base.des_cbfunc(&btl->super, frag->endpoint, &frag->base, OMPI_ERROR);
             OPAL_THREAD_LOCK(&mca_btl_gm_component.gm_lock);
+            if( btl_ownership ) {
+                mca_btl_gm_free(&btl->super, &frag->base);
+            }
             break;
     }
 }
@@ -806,12 +790,15 @@ static void mca_btl_gm_get_callback( struct gm_port* port, void* context, gm_sta
 {
     mca_btl_gm_frag_t* frag = (mca_btl_gm_frag_t*)context;
     mca_btl_gm_module_t* btl = frag->btl;
+    int btl_ownership = frag->base.des_flags & MCA_BTL_DES_FLAGS_BTL_OWNERSHIP;
 
     /* call the completion callback */
     switch(status) {
         case GM_TRY_AGAIN:
         case GM_SEND_TIMED_OUT:
+#if GM_API_VERSION > 0x200
         case GM_TIMED_OUT:
+#endif
             /* drop all sends to this destination port */
             gm_drop_sends(
                 btl->port,
@@ -837,6 +824,9 @@ static void mca_btl_gm_get_callback( struct gm_port* port, void* context, gm_sta
             OPAL_THREAD_UNLOCK(&mca_btl_gm_component.gm_lock);
             frag->base.des_cbfunc(&btl->super, frag->endpoint, &frag->base, OMPI_SUCCESS);
             OPAL_THREAD_LOCK(&mca_btl_gm_component.gm_lock);
+            if( btl_ownership ) {
+                mca_btl_gm_free(&btl->super, &frag->base);
+            }
 
             /* return the send token and deque pending fragments */
             MCA_BTL_GM_RETURN_TOKEN(btl);
@@ -852,6 +842,9 @@ static void mca_btl_gm_get_callback( struct gm_port* port, void* context, gm_sta
             OPAL_THREAD_UNLOCK(&mca_btl_gm_component.gm_lock);
             frag->base.des_cbfunc(&btl->super, frag->endpoint, &frag->base, OMPI_ERROR);
             OPAL_THREAD_LOCK(&mca_btl_gm_component.gm_lock);
+            if( btl_ownership ) {
+                mca_btl_gm_free(&btl->super, &frag->base);
+            }
             break;
     }
 }
@@ -972,3 +965,22 @@ int mca_btl_gm_finalize(struct mca_btl_base_module_t* btl)
     return OMPI_SUCCESS;
 }
 
+int mca_btl_gm_ft_event(int state) {
+    if(OPAL_CRS_CHECKPOINT == state) {
+        ;
+    }
+    else if(OPAL_CRS_CONTINUE == state) {
+        ;
+    }
+    else if(OPAL_CRS_RESTART == state) {
+        ;
+    }
+    else if(OPAL_CRS_TERM == state ) {
+        ;
+    }
+    else {
+        ;
+    }
+
+    return OMPI_SUCCESS;
+}
