@@ -1,14 +1,16 @@
+/* -*- Mode: C; c-basic-offset:4 ; -*- */
 /*
  * Copyright (c) 2004-2006 The Trustees of Indiana University and Indiana
  *                         University Research and Technology
  *                         Corporation.  All rights reserved.
- * Copyright (c) 2004-2005 The University of Tennessee and The University
+ * Copyright (c) 2004-2007 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
- * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
+ * Copyright (c) 2004-2007 High Performance Computing Center Stuttgart, 
  *                         University of Stuttgart.  All rights reserved.
  * Copyright (c) 2004-2005 The Regents of the University of California.
  *                         All rights reserved.
+ * Copyright (c) 2008      UT-Battelle, LLC
  * $COPYRIGHT$
  * 
  * Additional copyrights may follow
@@ -31,6 +33,11 @@
 #include "opal/class/opal_object.h"
 #include "ompi/mpi/f77/fint_2_int.h"
 
+#include <stdio.h>
+
+#if defined(c_plusplus) || defined(__cplusplus)
+extern "C" {
+#endif
 
 /**
  * Fortran handles; must be [manually set to be] equivalent to the
@@ -117,6 +124,8 @@ enum {
   /**< Floating point: double */
   OMPI_OP_TYPE_REAL,
   /**< Floating point: real */
+  OMPI_OP_TYPE_REAL2,
+  /**< Floating point: real*2 */
   OMPI_OP_TYPE_REAL4,
   /**< Floating point: real*4 */
   OMPI_OP_TYPE_REAL8,
@@ -185,12 +194,23 @@ enum {
  */
 typedef void (ompi_op_c_handler_fn_t)(void *, void *, int *, MPI_Datatype *);
 
+/*
+ *  Three buffer ( two input and one output) function prototype
+ */
+typedef void (ompi_op_3buff_c_handler_fn_t)(void * restrict , void * restrict, 
+        void * restrict, int *, MPI_Datatype *);
+
 
 /**
  * Typedef for fortran op functions.
  */
 typedef void (ompi_op_fortran_handler_fn_t)(void *, void *, 
                                             MPI_Fint *, MPI_Fint *);
+/*
+ * Three buffer (2 input one output) function prototype
+ */
+typedef void (ompi_op_3buff_fortran_handler_fn_t)(void * restrict, 
+        void * restrict , void * restrict , MPI_Fint *, MPI_Fint *);
 
 
 /**
@@ -203,6 +223,12 @@ typedef void (ompi_op_fortran_handler_fn_t)(void *, void *,
 typedef void (ompi_op_cxx_handler_fn_t)(void *, void *, int *, 
                                         MPI_Datatype *, MPI_User_function *op);
 
+/*
+ * Three buffer (two input, one output) function prototype
+ */
+typedef void (ompi_op_3buff_cxx_handler_fn_t)(void * restrict, void * restrict,
+        void * restrict, int *, MPI_Datatype *, MPI_User_function *op);
+
 
 /*
  * Flags for MPI_Op
@@ -256,12 +282,27 @@ struct ompi_op_t {
 
   /** Index in Fortran <-> C translation array */
   int o_f_to_c_index;
+
+  union {
+      /** C handler function pointer */
+      ompi_op_3buff_c_handler_fn_t *c_fn;
+      /** Fortran handler function pointer */
+      ompi_op_3buff_fortran_handler_fn_t *fort_fn;
+      /** C++ intercept function pointer -- see lengthy comment in
+          ompi/mpi/cxx/intercepts.cc::ompi_mpi_cxx_op_intercept() for
+          an explanation */
+      ompi_op_3buff_cxx_handler_fn_t *cxx_intercept_fn;
+  } o_3buff_func[OMPI_OP_TYPE_MAX];
+  /**< Array of three buffer function pointers, indexed on the operation 
+    type.  For non-intrinsice MPI_Op's, only the 0th element will be
+    meaningful. */
 };
+
 /**
  * Convenience typedef 
  */
 typedef struct ompi_op_t ompi_op_t;
-OBJ_CLASS_DECLARATION(ompi_op_t);
+OMPI_DECLSPEC OBJ_CLASS_DECLARATION(ompi_op_t);
 
 /**
  * Array to map ddt->id values to the corresponding position in the op
@@ -368,74 +409,65 @@ OMPI_DECLSPEC extern ompi_op_t ompi_mpi_op_replace;
 /**
  * Table for Fortran <-> C op handle conversion
  */
-OMPI_DECLSPEC extern struct ompi_pointer_array_t *ompi_op_f_to_c_table;
+extern struct opal_pointer_array_t *ompi_op_f_to_c_table;
 
+/**
+ * Initialize the op interface.
+ *
+ * @returns OMPI_SUCCESS Upon success
+ * @returns OMPI_ERROR Otherwise
+ *
+ * Invoked from ompi_mpi_init(); sets up the op interface, creates
+ * the predefined MPI operations, and creates the corresopnding F2C
+ * translation table.
+ */
+int ompi_op_init(void);
 
-#if defined(c_plusplus) || defined(__cplusplus)
-extern "C" {
-#endif
+/**
+ * Finalize the op interface.
+ *
+ * @returns OMPI_SUCCESS Always
+ *
+ * Invokes from ompi_mpi_finalize(); tears down the op interface, and
+ * destroys the F2C translation table.
+ */
+int ompi_op_finalize(void);
 
-  /**
-   * Initialize the op interface.
-   *
-   * @returns OMPI_SUCCESS Upon success
-   * @returns OMPI_ERROR Otherwise
-   *
-   * Invoked from ompi_mpi_init(); sets up the op interface, creates
-   * the predefined MPI operations, and creates the corresopnding F2C
-   * translation table.
-   */
-  int ompi_op_init(void);
+/**
+ * Create a ompi_op_t
+ *
+ * @param commute Boolean indicating whether the operation is
+ *        communative or not
+ * @param func Function pointer of the error handler
+ *
+ * @returns op Pointer to the ompi_op_t that will be
+ *   created and returned
+ *
+ * This function is called as the back-end of all the MPI_OP_CREATE
+ * functions.  It creates a new ompi_op_t object, initializes it to
+ * the correct object type, and sets the callback function on it.
+ *
+ * The type of the function pointer is (arbitrarily) the fortran
+ * function handler type.  Since this function has to accept 2
+ * different function pointer types (lest we have 2 different
+ * functions to create errhandlers), the fortran one was picked
+ * arbitrarily.  Note that (void*) is not sufficient because at
+ * least theoretically, a sizeof(void*) may not necessarily be the
+ * same as sizeof(void(*)).
+ *
+ * NOTE: It *always* sets the "fortran" flag to false.  The Fortran
+ * wrapper for MPI_OP_CREATE is expected to reset this flag to true
+ * manually.
+ */
+ompi_op_t *ompi_op_create(bool commute, ompi_op_fortran_handler_fn_t *func);
 
-  /**
-   * Finalize the op interface.
-   *
-   * @returns OMPI_SUCCESS Always
-   *
-   * Invokes from ompi_mpi_finalize(); tears down the op interface, and
-   * destroys the F2C translation table.
-   */
-  int ompi_op_finalize(void);
-
-  /**
-   * Create a ompi_op_t
-   *
-   * @param commute Boolean indicating whether the operation is
-   *        communative or not
-   * @param func Function pointer of the error handler
-   *
-   * @returns op Pointer to the ompi_op_t that will be
-   *   created and returned
-   *
-   * This function is called as the back-end of all the MPI_OP_CREATE
-   * functions.  It creates a new ompi_op_t object, initializes it to
-   * the correct object type, and sets the callback function on it.
-   *
-   * The type of the function pointer is (arbitrarily) the fortran
-   * function handler type.  Since this function has to accept 2
-   * different function pointer types (lest we have 2 different
-   * functions to create errhandlers), the fortran one was picked
-   * arbitrarily.  Note that (void*) is not sufficient because at
-   * least theoretically, a sizeof(void*) may not necessarily be the
-   * same as sizeof(void(*)).
-   *
-   * NOTE: It *always* sets the "fortran" flag to false.  The Fortran
-   * wrapper for MPI_OP_CREATE is expected to reset this flag to true
-   * manually.
-   */
-  ompi_op_t *ompi_op_create(bool commute, ompi_op_fortran_handler_fn_t *func);
-
-  /**
-   * Mark an MPI_Op as holding a C++ callback function, and cache
-   * that function in the MPI_Op.  See a lenghty comment in
-   * ompi/mpi/cxx/op.c::ompi_mpi_cxx_op_intercept() for a full
-   * expalantion.
-   */
-  void ompi_op_set_cxx_callback(ompi_op_t *op, MPI_User_function *fn);
-
-#if defined(c_plusplus) || defined(__cplusplus)
-}
-#endif
+/**
+ * Mark an MPI_Op as holding a C++ callback function, and cache
+ * that function in the MPI_Op.  See a lenghty comment in
+ * ompi/mpi/cxx/op.c::ompi_mpi_cxx_op_intercept() for a full
+ * expalantion.
+ */
+OMPI_DECLSPEC void ompi_op_set_cxx_callback(ompi_op_t *op, MPI_User_function *fn);
 
 /**
  * Check to see if an op is intrinsic.
@@ -521,15 +553,18 @@ static inline bool ompi_op_is_valid(ompi_op_t *op, ompi_datatype_t *ddt,
                   NULL == op->o_func[ompi_op_ddt_map[ddt->id]].fort_fn) ||
                  (0 == (op->o_flags & OMPI_OP_FLAGS_FORTRAN_FUNC) &&
                   NULL == op->o_func[ompi_op_ddt_map[ddt->id]].c_fn))) {
-                asprintf(msg, "%s: the reduction operation %s is not defined on the %s datatype", func, op->o_name, ddt->name);
+                asprintf(msg, "%s: the reduction operation %s is not defined on the %s datatype",
+                         func, op->o_name, ddt->name);
                 return false;
             }
         } else {
             /* Non-intrinsic ddt on intrinsic op */
             if ('\0' != ddt->name[0]) {
-                asprintf(msg, "%s: the reduction operation %s is not defined for non-intrinsic datatypes (attempted with datatype named \"%s\")", func, op->o_name, ddt->name);
+                asprintf(msg, "%s: the reduction operation %s is not defined for non-intrinsic datatypes (attempted with datatype named \"%s\")",
+                         func, op->o_name, ddt->name);
             } else {
-                asprintf(msg, "%s: the reduction operation %s is not defined for non-intrinsic datatypes", func, op->o_name);
+                asprintf(msg, "%s: the reduction operation %s is not defined for non-intrinsic datatypes",
+                         func, op->o_name);
             }
             return false;
         }
@@ -616,4 +651,91 @@ static inline void ompi_op_reduce(ompi_op_t *op, void *source, void *target,
   }
 }
 
+/**
+ * Perform a reduction operation.
+ *
+ * @param op The operation (IN)
+ * @param source Source1 (input) buffer (IN)
+ * @param source Source2 (input) buffer (IN)
+ * @param target Target (output) buffer (IN/OUT)
+ * @param count Number of elements (IN)
+ * @param dtype MPI datatype (IN)
+ *
+ * @returns void As with MPI user-defined reduction functions, there
+ * is no return code from this function.
+ *
+ * Perform a reduction operation with count elements of type dtype in
+ * the buffers source and target.  The target buffer obtains the
+ * result (i.e., the original values in the target buffer are reduced
+ * with the values in the source buffer and the result is stored in
+ * the target buffer).
+ *
+ * This function figures out which reduction operation function to
+ * invoke and whether to invoke it with C- or Fortran-style invocation
+ * methods.  If the op is intrinsic and has the operation defined for
+ * dtype, the appropriate back-end function will be invoked.
+ * Otherwise, the op is assumed to be a user op and the first function
+ * pointer in the op array will be used.
+ *
+ * NOTE: This function assumes that a correct combination will be
+ * given to it; it makes no provision for errors (in the name of
+ * optimization).  If you give it an intrinsic op with a datatype that
+ * is not defined to have that operation, it is likely to seg fault.
+ */
+static inline void ompi_3buff_op_reduce(ompi_op_t *op, void *source1, void *source2, 
+        void *target, int count, ompi_datatype_t *dtype)
+{
+  MPI_Fint f_dtype, f_count;
+  void * restrict src1;
+  void * restrict src2;
+  void * restrict tgt;
+  src1=source1;
+  src2=source2;
+  tgt=target;
+
+  /*
+   * Call the reduction function.  Two dimensions: a) if both the op
+   * and the datatype are intrinsic, we have a series of predefined
+   * functions for each datatype, b) if the op has a fortran callback
+   * function or not.
+   *
+   * NOTE: We assume here that we will get a valid result back from
+   * the ompi_op_ddt_map[] (and not -1) -- if we do, then the
+   * parameter check in the top-level MPI function should have caught
+   * it.  If we get -1 because the top-level parameter check is turned
+   * off, then it's an erroneous program and it's the user's fault.
+   * :-)
+   */
+
+  if (0 != (op->o_flags & OMPI_OP_FLAGS_INTRINSIC) &&
+     ompi_ddt_is_predefined(dtype)) {
+    if (0 != (op->o_flags & OMPI_OP_FLAGS_FORTRAN_FUNC)) {
+      f_dtype = OMPI_INT_2_FINT(dtype->d_f_to_c_index);
+      f_count = OMPI_INT_2_FINT(count);
+      op->o_3buff_func[ompi_op_ddt_map[dtype->id]].fort_fn(src1, src2 , tgt,
+                                                     &f_count, &f_dtype);
+    } else {
+      op->o_3buff_func[ompi_op_ddt_map[dtype->id]].c_fn(src1, src2, tgt,&count,
+                                                  &dtype);
+    }
+  } 
+
+  /* User-defined function - this can't work, will never be called.
+   *  need to take this out soon. */
+
+  else if (0 != (op->o_flags & OMPI_OP_FLAGS_FORTRAN_FUNC)) {
+    f_dtype = OMPI_INT_2_FINT(dtype->d_f_to_c_index);
+    f_count = OMPI_INT_2_FINT(count);
+    op->o_3buff_func[0].fort_fn(src1, src2, tgt, &f_count, &f_dtype);
+  } else if (0 != (op->o_flags & OMPI_OP_FLAGS_CXX_FUNC)) {
+    op->o_3buff_func[0].cxx_intercept_fn(src1, src2, tgt, &count, &dtype,
+                                   op->o_func[1].c_fn);
+  } else {
+    op->o_3buff_func[0].c_fn(src1, src2, tgt, &count, &dtype);
+  }
+}
+#if defined(c_plusplus) || defined(__cplusplus)
+}
+#endif
+
 #endif /* OMPI_OP_H */
